import AutoRPE.UtilsRPE.VariablePrecision as VariablePrecision
import AutoRPE.UtilsRPE.BasicFunctions as BasicFunctions
import AutoRPE.UtilsRPE.ModifyCodeUtils as ModifyCodeUtils
import AutoRPE.UtilsRPE.Vault as Vault


def change_code(sources_path, var_to_be_changed, target_precision):
    """
    The purpose of this function is to modify the declarations of several variables in a source-code.

    - source_path: Path to a folder that contains the source files
    - var_to_be_changed: list of variables that will be modified in the code
    - vault_path: Path to the database that contains info about source files
    """

    # Load files that will be modified
    modules = BasicFunctions.load_sources(sources_path, extension=[".F90", ".h90", ".f90"])

    # For every variable in the banned, searches declaration through source files and change declaration's precision
    for index, variable in enumerate(var_to_be_changed):
        ModifyCodeUtils.update_precision_declaration(variable, modules, target_precision)
        print("Updated precision of var ", variable.name, index)

    # Save file to disk overwriting originals (this way preserve the folder structure, too difficult to reproduce)
    for module in modules:
        module.save_file()


def load_variable_list(filename):
    # Load list of variables
    from numpy import loadtxt
    list_of_variables = loadtxt(filename, dtype=int).tolist()
    if not isinstance(list_of_variables, list):
        list_of_variables = [list_of_variables]
    return list_of_variables


def get_command_line_arguments():
    import argparse
    from AutoRPE.UtilsRPE.Error import PathType
    parser = argparse.ArgumentParser()
    parser.add_argument("--input-files", required=True, type=PathType(exists=True, type="dir"), nargs='+',
                        help="Path to the non preprocessed sources")
    parser.add_argument("--variable-list", required=True, type=PathType(exists=True, type='file'),
                        help="Path to the filename with the list of variables")
    parser.add_argument("--vault", required=True, type=PathType(exists=True, type='file'),
                        help="Path to vault generated either from RPE implementation")
    parser.add_argument("-variable-precision", default="dp",
                        help="Precision specification at which the variables in the list will be set")
    parser.add_argument("--working-precision", default="23", type=int,
                        help="Precision used for wp variables in the code. i.e. double for standard runs, "
                             "single for mixed-precision versions")

    return parser.parse_args()


if __name__ == "__main__":
    args = get_command_line_arguments()

    input_files = args.input_files
    variable_list = args.variable_list
    vault_path = args.vault
    variable_precision = args.variable_precision
    working_precision = args.working_precision

    # Set working precision
    VariablePrecision.set_working_precision(working_precision)

    # Load list of variable ids
    variable_id = load_variable_list(variable_list)
    vault = Vault.load_vault(vault_path)

    # Get variables from their ids
    variables = [vault.get_variable_by_id(identifier) for identifier in variable_id]

    # List variables used in external calls, because they have to be change precision too
    used_var_in_ext = [v for v in vault.variables if v.is_used_in_external and v.type == "rpe_var"]
    variables += used_var_in_ext

    # Propagate dependencies of the variables using same_as property
    for var in variables:
        for additional_var in var.same_as:
            # Only if the variable has an id ban it
            if additional_var not in variables:
                variables.append(additional_var)

    # Keep unique values
    all_variables = list(set(variables))

    # This is for debugging purposes; from a run to the next we want errors to happen always in the same order
    all_variables.sort(key=lambda x: x.incremental_number)

    # Change precision of variables in source code
    change_code(input_files, all_variables, variable_precision)
