import AutoRPE.UtilsRPE.BasicFunctions as BasicFunctions
import AutoRPE.UtilsRPE.BasicStructures as BasicStructures
import AutoRPE.UtilsRPE.CallManager as CallManager
import AutoRPE.UtilsRPE.Inserter as Inserter
import AutoRPE.UtilsRPE.Splitter as Splitter
import AutoRPE.UtilsRPE.AttributeParser as AttributeParser
import AutoRPE.UtilsRPE.VariablePrecision as VariablePrecision
import AutoRPE.UtilsRPE.RegexPattern as RegexPattern
import AutoRPE.UtilsRPE.Error as Error
import re


def update_precision_declaration(variable, modules, target_precision="dp"):
    """
    Find the file where a variable is declared and replace the declaration precision specification with target_precision
    """
    var_module = variable.procedure.module.name
    # This is the module where the variables appear declared once the code is preprocessed
    preprocessed_module = [m for m in modules if m.module_name == var_module]

    if len(preprocessed_module) > 1 or not len(preprocessed_module):
        raise Error.ExceptionNotManaged

    preprocessed_module = preprocessed_module[0]
    included_files = find_included_files(preprocessed_module, modules)

    # Find the variable original module and line declaration, merge the line if is split
    if variable.is_retval:
        print("The variable %s is actually a function. Skip" % variable.name)
        return

    original_module, index_to_modify = find_and_merge_variable_declaration_line(variable, preprocessed_module,
                                                                                included_files)
    line_to_modify = original_module.lines[index_to_modify]

    # Check that precision declaration has a different precision wrt target_precision
    original_precision = re.search(RegexPattern.precision_of_real_declaration, line_to_modify, re.I)

    if not original_precision:
        raise Error.ExceptionNotManaged("Precision was not found for this declaration")
    original_precision = original_precision.group(3)
    # Check the precision is correct
    try:
        VariablePrecision.lookup_table[original_precision]
    except KeyError:
        raise Error.ExceptionNotManaged("Something went wrong, %s is not a real precision" % original_precision)

    # Avoid replacing variables whose precision was already of the correct type
    # TODO: these variables should not enter the analysis from the beginning
    if original_precision == target_precision:
        return

    original_module.lines[index_to_modify] = replace_real_precision(variable, line_to_modify,
                                                                    target_precision=target_precision)
    original_module.rebuild_text()


def replace_real_precision(variable, declaration_line, target_precision="dp"):
    """This function takes a declaration line (string) and updates it with the new type
    If the declaration contains more variables it keeps the other variables with the original attributes."""
    import warnings
    declaration_line_without_comments = BasicFunctions.remove_comments(declaration_line).strip()

    attributes, arguments = declaration_line_without_comments.split("::")
    variable_name = AttributeParser.get_variable_name(arguments)

    # Real statement pattern precision_of_real_declaration
    pattern = "real *\(([^\)]*)\)"

    if len(variable_name) == 1:
        m = re.search(pattern, declaration_line_without_comments, re.I)
        if m:
            real_statement = m.group(0)
            precision_specification = m.group(1)
            new_real_statement = real_statement.replace(precision_specification, target_precision)
            new_declaration_line_without_comments = \
                declaration_line_without_comments.replace(real_statement, new_real_statement)
            new_declaration_line = \
                declaration_line.replace(declaration_line_without_comments, new_declaration_line_without_comments)
        else:
            warnings.warn("Problems finding real statement:\n%s" % declaration_line)
            new_declaration_line = "! WARNING!! the declaration hasn't been properly fixed!\n" + \
                                   declaration_line
        return new_declaration_line
    else:
        m = re.search(pattern, attributes, re.I)
        if m:
            real_statement = m.group(0)
            precision_specification = m.group(1)
            new_real_statement = real_statement.replace(precision_specification, target_precision)
            new_attributes = attributes.replace(real_statement, new_real_statement)
            additional_declaration = "%s :: %s" % (new_attributes, variable.name)
            m = re.search(r"(.*\b)%s(\b.*)" % variable.name, declaration_line_without_comments)
            if m:

                declaration_line_without_selected_var = "".join(m.groups(0))
            else:
                warnings.warn("WTF?")

            # Comma patterns to remove the extra comma
            beginning_pat = "::( *\,)"
            beginning_pat_substitution = "::"
            middle_pat = ",( *,)"
            middle_pat_sub = ","
            last_pat = "(, *)$"
            last_pat_sub = ""

            for pat, sub in [(beginning_pat, beginning_pat_substitution), (middle_pat, middle_pat_sub),
                             (last_pat, last_pat_sub)]:
                m = re.search(pat, declaration_line_without_selected_var)
                if m:
                    new_declaration_line_without_comments = re.sub(pat, sub, declaration_line_without_selected_var)
                    new_declaration_line = declaration_line.replace(declaration_line_without_comments,
                                                                    new_declaration_line_without_comments)
                    break
        else:
            pass

        # Adding proper indentation
        leading_spaces = (len(new_declaration_line) - len(new_declaration_line.lstrip()))
        new_declaration_line = "%s%s\n%s" % (
            leading_spaces * " ", additional_declaration.lstrip(), new_declaration_line)

        return new_declaration_line


def find_and_merge_variable_declaration_line(variable, module, additional_modules):
    """
    Given a variable and a module:
              1) Search for the variable.subprogram declaration
              2) Among the declarations search the variable.name
              3) If the variable declaration is split, merge the line and update both the module and the index
    Returns the index
    """

    # The var is a global variable, declared at the beginning of a module
    if variable.procedure.name == "main":
        start, end = find_main_declarations_index(module)
    # The variable is member of a structure
    elif isinstance(variable.procedure, BasicStructures.DerivedType):
        start, end = find_structure_declarations_index(variable.procedure, module)
    # The variable is a private var of a subprogram
    else:
        try:
            start, end = find_subprogram_declarations_index(variable.procedure, module)
        except Error.ProcedureNotFound:
            module, start, end = search_subprogram_in_additional_modules(variable.procedure, additional_modules)

    declaration_index = None
    for index, line in enumerate(module.lines[start:end + 1]):
        # Skip lines that contains just comments
        line = BasicFunctions.remove_comments(line)
        if not line.strip():
            continue
        # The firs match will be declaration, since we are checking only declaration lines
        if re.search(RegexPattern.escape_string(variable.name), line):
            # Check declaration is on the same line of precision specification
            declaration_index = start + index
            if not line.count('::') or line.count("&"):
                # It rebuilds split line inside the module object and return merged line, and updated index
                declaration_index = rebuild_split_line(declaration_index, module)
            break

    if declaration_index is None:
        # The var can be declared in some .h90 included file
        if not additional_modules:
            raise Error.ExceptionNotManaged("Check for variables in include files, probably recursively")
        for additional_mod in additional_modules:
            module, declaration_index
        raise Error.ExceptionNotManaged("Line not found")

    # Return original module object  and index of the declaration
    return module, declaration_index


def search_subprogram_in_additional_modules(procedure, additional_modules):
    if not len(additional_modules):
        raise Error.ProcedureNotFound
    for additional_m in additional_modules:
        try:
            start, end = find_subprogram_declarations_index(procedure, additional_m)
        except Error.ProcedureNotFound:
            continue
        return additional_m, start, end
    raise Error.ProcedureNotFound


def find_main_declarations_index(module):
    """Find start and end indexes of variables declaration in main """
    main_start = -1
    # In case in main there is no subprogram declaration, just variables
    main_end = len(module.lines) - 1
    for index, line in enumerate(module.lines):
        line = BasicFunctions.remove_comments(line)
        # Start of the declaration section
        if line.count("::") and main_start < 0:
            main_start = index
        # Subprogram declaration start
        if line.lower().count('contains'):
            main_end = index - 1
            break
    # If no variable declaration was found
    if main_start < 0:
        return [-1, -1]
    return [main_start, main_end]


def find_structure_declarations_index(structure, module):
    """Find start and end indexes of variables declaration of custom data type """
    start = -1

    begin_pattern = RegexPattern.type_name_begin % structure.name
    end_pattern = RegexPattern.end_type_statement
    for index, line in enumerate(module.lines):
        line = BasicFunctions.remove_comments(line)
        # Start of the declaration section, just if is has not been found yet (avoid match with the end pattern)
        if start < 0 and re.search(begin_pattern, line, re.I):
            start = index
        # Subprogram declaration start
        if start > 0 and re.search(end_pattern, line, re.I):
            end = index - 1
            break
    # Structure was not found
    if start < 0:
        return [-1, -1]
    return [start, end]


def find_subprogram_lines(subprogram, module):
    indices = []
    # if module is None:
    #     module = subprogram.module
    begin_pattern = RegexPattern.subprogram_name_declaration % subprogram.name
    end_pattern = RegexPattern.subprogram_name_end_declaration % subprogram.name
    # First find the subprogram body inside the module
    for index, line in enumerate(module.lines):
        # If it's the start or end of the subprogram
        line = BasicFunctions.remove_comments(line)
        if re.search(begin_pattern, line, re.I):
            indices.append(index)
            # End of the program was hit
            if re.search(end_pattern, line, re.I):
                break
    if not len(indices):
        raise Error.ProcedureNotFound
    # In case more than one start/end index was found:
    #
    #   #if defined key_agrif
    #      RECURSIVE SUBROUTINE stp_MLF( )
    #         INTEGER             ::   kstp   ! ocean time-step index
    #   #else
    #      SUBROUTINE stp_MLF( kstp )
    #         INTEGER, INTENT(in) ::   kstp   ! ocean time-step index
    #   #endif
    #
    # return the broather range
    return indices[0], indices[-1]


def find_subprogram_declarations_index(subprogram, module):
    """Find start and end indexes of variables declaration in subprogram code """
    # Find start and end of subprogram declaration
    indices = find_subprogram_lines(subprogram, module)

    start = -1
    end = 0
    # The search for declaration
    for index, line in enumerate(module.lines[indices[0]:indices[1] + 1]):

        line = BasicFunctions.remove_comments(line)
        # Skip only comment lines
        if not line.strip():
            continue
        # Skip until first declaration is found
        if start < 0:
            if not line.count('::'):
                continue
        # First declaration found
        if start < 0:
            start = indices[0] + index

        # There can be anything but comments among var declarations (but they can be split)
        # if is not a declaration means the body of the function begun
        if not (line.count('::') or previous_line_is_split(indices[0] + index, module.lines)) and start > 0:
            end = indices[0] + index - 1
            break

    # Return variables declaration line indexes
    return start, end


def add_preprocessor_cast(var_name, precision, original_module, original_line):
    if precision == "wp":
        cast = "CASTWP(%s)"
    elif precision == "sp":
        cast = "CASTSP(%s)"
    elif precision == "dp":
        cast = "CASTDP(%s)"
    else:
        raise Error.ExceptionNotManaged("Wrong input precision")
    for index, line in enumerate(original_module.lines[original_line.idx: original_line.idx +
                                                                          original_line.n_of_merged_lines]):
        # Avoid casting more than once but check if on another split line the variable is present
        if line.count(cast % var_name):
            purged_line = BasicFunctions.remove_literals(line)
            if purged_line.count(cast % var_name) != purged_line.count(var_name):
                raise Error.ExceptionNotManaged("Line was not fixed properly: check")
            continue
        if line.count(var_name):
            assignation = re.search(RegexPattern.assignation, line, re.I)
            if assignation:
                new_line = Inserter.replace_variable_exact_match(var_name, cast % var_name, assignation.group(3))
                new_line = assignation.group(1) + new_line
            else:
                new_line = Inserter.replace_variable_exact_match(var_name, cast % var_name, line)
            return index + original_line.idx, new_line
    return index + original_line.idx, line


def find_subprogram_call(subprogram_call, argument_index, original_modules):
    module = subprogram_call.block.module
    # Search just for the first one to fix
    argument_names = [subprogram_call.arguments[index].name for index in argument_index]
    # Find the original module
    original_module = [om for om in original_modules if om.module_name == module.name][0]
    # Find start and end of subprogram block where the fix is due
    indices = find_subprogram_lines(subprogram_call.block, original_module)
    # Name of the function to be searched
    if subprogram_call.subprogram.interface:
        subprogram_name = subprogram_call.subprogram.interface.name
    else:
        subprogram_name = subprogram_call.subprogram.name
    if isinstance(subprogram_call.subprogram, BasicStructures.Function):
        subprogram_pattern = RegexPattern.call_to_function_name
    elif isinstance(subprogram_call.subprogram, BasicStructures.SubRoutine):
        subprogram_pattern = RegexPattern.call_to_subroutine_name
    else:
        raise Error.ExceptionNotManaged("Something unexpected ended up here!")
    # Initialize the object to be filled within the loop
    found_lines = FoundLines(original_module)
    # Search for all the calls with this program name and argument inside the subprogram body
    for index, line in enumerate(original_module.lines[indices[0]:indices[1] + 1]):
        # Return merged line if the line is split
        line = BasicFunctions.merge_line_range(original_module.lines, indices[0] + index)
        line = BasicFunctions.remove_comments(line)
        line = BasicFunctions.remove_if_else_condition(line)
        # Avoid matching  when the variable is on one side of the equation and the function on the other
        assignation = re.search(RegexPattern.assignation, line, re.I)
        if assignation:
            line = assignation.group(3)
        if re.search(subprogram_pattern % subprogram_name, line, re.I):
            # The function is found: check is a call with all the arguments to fix
            arg_found = [ag for ag in argument_names if re.search(RegexPattern.escape_string(ag), line, re.I)]
            if len(arg_found) == len(argument_names):
                line_obj = Lines(indices[0] + index, BasicFunctions.merge_line_range.number)
                found_lines.lines.append(line_obj)
        # if line was split, jump the merged indices
        for i in range(BasicFunctions.merge_line_range.number - 1):
            continue
    if not found_lines:
        raise Error.LineNotFound("Something went wrong")
    return found_lines


def previous_line_is_split(index, module_lines):
    """Check if last previous line that was not a comment was a split line"""
    index -= 1
    while not BasicFunctions.remove_comments(module_lines[index]).strip():
        index -= 1
    if module_lines[index].count("&"):
        return True
    return False


def rebuild_split_line(line_index, original_module):
    """Takes a line that can be either split upward or downward, rebuilt it and modify the original module text"""
    # The variable is declared on the first line, but the line is split downward
    if original_module.lines[line_index].count("::"):
        merge_line_in_file(line_index, original_module)
    # The variable is one or more lines down the declaration, go up until the declaration is found
    while not original_module.lines[line_index].count("::"):
        line_index -= 1
    # Now merge lines downwards
    merge_line_in_file(line_index, original_module)
    return line_index


def merge_line_in_file(line_index, module):
    """
    Given a line index corresponding to a split line, copies all the  pieces up to the line index and remove empty lines
    """
    line = BasicFunctions.merge_line_range(module.lines, line_index)
    module.lines[line_index] = line
    # Remove lines that will be empty now
    for j in range(1, BasicFunctions.merge_line_range.number):
        module.lines[line_index + j] = ""
    # Rebuilt text and lines
    module.rebuild_text()


def find_included_files(original_module, modules):
    # The var can be declared in some .h90 included file
    additional_modules = []
    # Search for the include pattern and return the h90 file name
    additional_module_names = [re.search(r"^#\s*include\s+\"(\w+.h90)\"", l, re.I).group(1)
                               for l in original_module.lines if re.search(r"^#\s*include", l, re.I)]
    # If some h90 file is included, create a list with corresponding modules object
    if additional_module_names:
        additional_modules = [m for m in modules if m.filename in additional_module_names]
    return additional_modules


def get_array_of_interfaces(list_of_interface):
    if len(list_of_interface) > 1:
        # Reverse sort the list from longest to shortest
        list_of_interface.sort(key=len)
        list_of_interface = list_of_interface[::-1]
        # Save the longest, that will serve to check the others are not already present in this one
        array_of_interfaces = [list_of_interface[0]]
        # Check the rest
        for call in list_of_interface[0:]:
            match = False
            for i, arg in enumerate(call):
                if not match:
                    for interf in array_of_interfaces:
                        if arg not in interf:
                            continue
                        if i + 1 == len(call):
                            match = True
                            break
            if not match:
                # This interface is not yet in the list
                array_of_interfaces.append(call)
    else:
        array_of_interfaces = [list_of_interface[0]]
    return array_of_interfaces


def arguments_to_cast(subprogram_call):
    ret_val = []
    for s_call in subprogram_call:
        if not s_call.subprogram.is_external:
            for arg, dummy_a in zip(s_call.arguments, s_call.dummy_arguments):
                # If dummy_a is of the same type of argument
                # or one is wp and the other is dp:return True if wp=dp, False otherwise
                if VariablePrecision.type_are_equivalent(arg.type, dummy_a.type, working_precision="sp"):
                    continue
                # If the called arg is external and the dummy is not real, no need to check
                if arg.type == "external" and dummy_a.type not in VariablePrecision.real_id:
                    continue
                # Inconsistency between used argument and what the call expects
                if dummy_a.intent == "in":
                    # If is wp, maintain wp
                    # If is sp for sure there is a dp version of this, otherwise the code would not work in dp
                    if  dummy_a.type in [VariablePrecision.real_id['sp'], VariablePrecision.real_id['wp']]:
                        cast_precision = "wp"
                    else:
                        # There is no sp/dp interface (otherwise we would be in the other branch of the if/else)
                        cast_precision = "dp"
                    # From the dummy argument name in the subprogram object get its real position in this call
                    dummy_argument_name = dummy_a.name
                    dummy_index_in_call = [da.name for da in s_call.dummy_arguments].index(dummy_argument_name)
                    # If a call that need fix to more than on arg, don't create another object, just append the index
                    call = [rv for rv in ret_val if rv.subprogram_call == s_call]
                    if call:
                        call[0].argument_index += [dummy_index_in_call]
                    else:
                        # This is the first argument on this call to get a fix: create the object
                        subprogram_cast = SubprogramCast(s_call, dummy_index_in_call, cast_precision)
                        ret_val.append(subprogram_cast)

                # The value of the argument can be modified by the call
                else:
                    print("This should not happen! Check the modify variable script %s" % s_call.call)
                    # raise Error.ExceptionNotManaged("This should not happen! Check the modify variable script")
        else:
            # This is only temporal, we don't need to check external calls, they should be already ok
            called_arguments = s_call.arguments
            # In case the procedure is not found, consider it external
            for argument in called_arguments:
                # This is a stupid workaround, since it works for NEMO but possibly not with other codes
                if argument.type == "sp" and not s_call.block.name.count("_sp"):
                    print("\nThis should not happen! Check the modify variable script %s" % s_call.call)
    return ret_val


class SubprogramCast:
    def __init__(self, subprogram_call, index, precision):
        self.subprogram_call = subprogram_call
        self.argument_index = index if isinstance(index, list) else [index]
        self.precision = precision


class Lines:
    def __init__(self, line_index, n_of_merged_lines):
        self.idx = line_index
        self.n_of_merged_lines = n_of_merged_lines


class FoundLines:
    def __init__(self, original_module):
        self.original_module = original_module
        self.lines = []


def add_include_clause(module):
    contains_marker = [index for index, l in enumerate(module.lines) if l.lower().count("contains")]
    contains_marker = contains_marker[0]
    include_lines = [index for index, l in enumerate(module.lines[:contains_marker]) if re.search(r"#\s+include ", l)]
    already_present = [idx for idx in include_lines if re.search("single_precision_substitute.h90", module.lines[idx])]
    if already_present:
        return
    if include_lines:
        include_lines = include_lines[-1]
    else:
        include_lines = contains_marker
    module.lines[include_lines] = "#  include \"single_precision_substitute.h90\"" + "\n" + module.lines[include_lines]


def get_reshape_fixes(vault, dictionary_of_fix):
    n_of_modules = len(vault.modules)
    subprogram = BasicStructures.Function('RESHAPE', None, is_external=True)
    for index, module in enumerate(vault.modules):
        print("\rFinding reshape exception %i/%i  %20s" % (index + 1, n_of_modules, module.name), end="")
        for line_index, line in enumerate(module.lines):
            # Returns arguments of subprograms call and corresponding type and dummy
            temporal_call = BasicFunctions.remove_if_condition(line)
            reshape_call = re.search(RegexPattern.reshape_function, temporal_call, re.I)
            if reshape_call:
                # for call like RESHAPE((/(zshift + REAL(ji,wp), ji = 1, jpi_crs*jpj_crs) /), (/ jpi_crs, jpj_crs /))
                if reshape_call.group(1).count("="):
                    continue
                else:
                    called_arguments = Splitter.split_hardcoded_array(reshape_call.group(1))
                call = reshape_call.group(0)
                subprogram_call = CallManager.create_SubprogramCall(call,
                                                                    called_arguments,
                                                                    subprogram, module.blocks[line_index], vault)
                # If something was found check if can be fixed
                different_type = list(set([a.type for a in subprogram_call.arguments]))
                if len(different_type) != 1:
                    type_1_idx = [idx for idx, a in enumerate(subprogram_call.arguments) if a.type == different_type[0]]
                    type_2_idx = [idx for idx, a in enumerate(subprogram_call.arguments) if a.type == different_type[1]]

                    if len(type_1_idx) > len(type_2_idx):
                        subprogram_cast = SubprogramCast(subprogram_call, type_2_idx, different_type[0])
                    else:
                        subprogram_cast = SubprogramCast(subprogram_call, type_1_idx, different_type[1])

                    try:
                        dictionary_of_fix["RESHAPE"] += [subprogram_cast]
                    except KeyError:
                        dictionary_of_fix["RESHAPE"] = [subprogram_cast]


def get_subprogram_fixes(vault, dictionary_of_fix):
    n_of_modules = len(vault.modules)
    for index, module in enumerate(vault.modules):
        print("\rFinding subprogram exception %i/%i  %20s" % (index + 1, n_of_modules, module.name), end="")
        for line_index, line in enumerate(module.lines):
            # Returns arguments of subprograms call and corresponding type and dummy
            temporal_call = BasicFunctions.remove_if_condition(line)
            try:
                subprogram_call = \
                    CallManager.get_procedure_inline(temporal_call, vault, module.blocks[line_index])
            except Error.InterfaceNotFound:
                # The interface does not exist for dp/sp variables
                VariablePrecision.set_working_precision('dp')
                subprogram_call = \
                    CallManager.get_procedure_inline(temporal_call, vault, module.blocks[line_index])
                VariablePrecision.set_working_precision('sp')
            # If something was found check if can be fixed
            if subprogram_call:
                subprogram_cast = arguments_to_cast(subprogram_call)
                if subprogram_cast:
                    for spb_cast in subprogram_cast:
                        if spb_cast.subprogram_call.subprogram.interface:
                            # If it is an interface, we have to search for the generic name, not the interface
                            name = spb_cast.subprogram_call.subprogram.interface.name
                        else:
                            name = spb_cast.subprogram_call.subprogram.name
                        try:
                            dictionary_of_fix[name] += [spb_cast]
                        except KeyError:
                            dictionary_of_fix[name] = [spb_cast]


def get_fixes(vault):
    # Create the dictionary
    dictionary_of_fix = {}
    # Fill dictionary with subprogram that need fix
    get_subprogram_fixes(vault, dictionary_of_fix)
    # Fill dictionary with reshape call that need fix
    get_reshape_fixes(vault, dictionary_of_fix)

    return dictionary_of_fix


def add_cast(original_input_files, dictionary_of_fix):
    original_modules = BasicFunctions.load_sources(original_input_files, extension=[".f90", ".F90", ".h90"])
    print("\n")
    len_of_fix = len(dictionary_of_fix)
    for index, subprogram_name in enumerate(dictionary_of_fix):
        print("\rFixing exception %i/%i" % (index + 1, len_of_fix), end="")
        subprogram_cast = dictionary_of_fix[subprogram_name]

        for call in subprogram_cast:
            found_lines = find_subprogram_call(call.subprogram_call, call.argument_index,
                                               original_modules=original_modules)
            if not found_lines.lines:
                raise Error.LineNotFound("There was some problem with the call %" % call.subprogram_call.call)
            for line in found_lines.lines:
                original_module = found_lines.original_module
                add_include_clause(original_module)

                # Fix all the argument that need cast in line
                for ind in call.argument_index:
                    var_to_fix = call.subprogram_call.arguments[ind].name
                    index, fixed_line = add_preprocessor_cast(var_to_fix, call.precision, original_module, line)
                    original_module.lines[index] = fixed_line
    return original_modules
