######################################
#
#
# This class was originally created as a simple queue, a place to add the jobs that had to be executed.
# However, it became convenient to add several methods to organize the execution of the jobs.
#


class ExperimentDriver(object):
    """
    The objects are initialized providing a seed job, i.e. a job that will be used to start a binary search.
    This job needs to be a BinarySearchJob object.
    Also, the maximum number of running jobs can be modified.

    # This object basically consists of several lists containing the jobs that are part of this experiment.
    # We will have a list for each one of the possible statuses, that are the following:
        # PENDING: The job has been created and ready to be launched.
        # RUNNING: The job is RUNNING.
        # SUCCESS: A job that has been completed with successful results and asserted.
        # FAILED: A job that has been completed with unsuccessful results and that can not be split.
        # SUSPENDED: A job that has been completed with unsuccessful results that has been split and depends on the
        # results of descendant jobs.
    """
    def __init__(self, seed_job=None, max_running_jobs=100):

        self.loop = 0
        # Initialize the different lists
        self.pending = []
        self.running = []
        self.success = []
        self.failed = []
        self.suspended = []
        self.disinherited = []
        # Initialize a dictionary with links to all the lists
        self.__dict = {"PENDING": self.pending,
                       "RUNNING": self.running,
                       "SUCCESS": self.success,
                       "FAILED": self.failed,
                       "SUSPENDED": self.suspended,
                       "DISINHERITED": self.disinherited,
                       }
        # Save the seed job
        self.seed_job = seed_job
        # Set the number of maximum jobs running at the same time
        self.max_running_jobs = max_running_jobs

        # Set checkpoint name and put seed job to pending
        if self.seed_job is not None:
            self.pending.append(self.seed_job)
            self.checkpoint_name = "checkpoint_%i.pkl" % self.__hash__()

        # Not sure about if this can happen
        else:
            self.checkpoint_name = "checkpoint.pkl"

    def sort(self, lower_first=True):
        """
        Depending on the situation it might be interesting to prioritize jobs of lower or higher level.
        Not very useful right now.
        """
        if lower_first:
            self.pending.sort(key=lambda _x: _x.level)
        else:
            self.pending.sort(key=lambda _x: - _x.level)

    def all(self):
        """
        :return: Returns all the jobs contained in all the different lists (excluding disinherited)
        """
        return self.running + self.pending + self.success + self.failed + self.suspended

    def __getitem__(self, item):
        """
        Define getitem method to enable to directly access the dictionary without having to specify it.
        :return: A given list corresponding to the item provided
        """
        return self.__dict[item]

    def step(self):
        """
        The function will loop through the jobs in the queue and will execute the driver.
        :return: Number of jobs that changed status
        """
        changes = 0
        for job in self.all():
            if job.driver(queue=self, max_running_jobs=self.max_running_jobs):
                changes += 1
                self.seed_job.graph(self.disinherited)

        self.loop += changes
        return changes

    def print_status(self):
        """
        Small utility to print how many jobs there are in which list.
        Useful to know the progress of the experiment.
        """
        for key in self.__dict.keys():
            print("\t%18s: %4i" % (key, len(self[key])))

    # def run(self, condition):
    #     """
    #     Start running the experiment and don't exit until the condition provided is fulfilled.
    #     """
    #     import time
    #
    #     loops = 0
    #     change_counter = 0
    #     wait = False
    #     waiting_time = 10
    #     while condition(self.seed_job):
    #         # Trying to avoid too much ssh polling by waiting if the loop time is less than a threshold
    #         loop_start_time = time.time()
    #         # Main loop
    #         print("ITERATION: %4i" % loops)
    #         self.print_status()
    #         # Sorting the queue to launch first the lower levels (bigger groups).
    #         self.sort(lower_first=True)
    #         # Launch one step (will loop through the jobs in the queue and take actions according to their status)
    #         change_counter += self.step()
    #         if change_counter > 10:
    #             self.checkpoint()
    #             change_counter = 0
    #         loop_end_time = time.time()
    #         if wait:
    #             while loop_end_time - loop_start_time < waiting_time:
    #                 time.sleep(1)
    #                 loop_end_time = time.time()
    #     print("done")

    def checkpoint(self, filename=None):
        """
        Save checkpoint with the current status of the experiment.
        """
        from AutoRPE.UtilsWorkflow.CheckpointTools import checkpoint
        if filename is None:
            filename = self.checkpoint_name
        checkpoint(queue=self, iteration=self.loop, filename=filename)

    def __hash__(self):
        """
        The hash of the experiment will be the same as the hash of the seed job.
        """
        return self.seed_job.hash